@@ -2,23 +2,23 @@
 
 import base64
 import collections
-from re import template
 import glog
 from google.protobuf import descriptor
 from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
 from google.protobuf.message import Message as ProtoMessage
 import jinja2
-import networkx as nx
 import os
 import pathlib
-from typing import Dict, Iterable, List, Optional, Tuple
+from typing import List, Optional, Tuple
 
 from ascend.protos.ascend import ascend_pb2
 from ascend.protos.schema import schema_pb2
+from ascend.sdk.connection_value import Union as UnionValue
+from ascend.sdk.connection_value import value_from_proto
 from ascend.sdk.builder import dataflow_from_proto, data_service_from_proto
 from ascend.sdk.client import Client
 from ascend.sdk.common import components_ordered_by_dependency, dataflows_ordered_by_dependency
-from ascend.sdk.definitions import Component, Dataflow, DataService, Definition, ReadConnector, Transform
+from ascend.sdk.definitions import Component, Dataflow, DataService, Definition, ReadConnector, Transform, WriteConnector
 
 # A bit funky of an approach to support multiple versions of protobuf that use different internal container classes
 REPEATED_CONTAINERS = [RepeatedCompositeFieldContainer]
@@ -237,15 +237,13 @@ def _classname_map() -> dict:
 def _inline_code_list_for(component: Component, resource_path_prefix: str, base_path: str) -> Optional[List[InlineCode]]:
   if isinstance(component, Transform) and component.operator:
     if component.operator.HasField("spark_function") and component.operator.spark_function.executable.code.source.HasField("inline"):
-      if component.operator.spark_function.executable.code.language.HasField("sql") or \
-        component.operator.spark_function.executable.code.language.HasField("snowflake_sql") or \
-        component.operator.spark_function.executable.code.language.HasField("bigquery_sql"):
+      language = component.operator.spark_function.executable.code.language.WhichOneof("language")
+      if language in ("sql", "snowflake_sql", "databricks_sql", "bigquery_sql"):
         resource_name = f"{component.id}.sql"
-      elif component.operator.spark_function.executable.code.language.HasField("python"):
+      elif language in ("python", "snowpark"):
         resource_name = f"{component.id}.py"
       else:
-        l = component.operator.spark_function.executable.code.language.WhichOneof("language")
-        raise ValueError(f"code download for language type \"{l}\" is unsupported")
+        raise ValueError(f"code download for language type \"{language}\" is unsupported")
       return [
           InlineCode(code=component.operator.spark_function.executable.code.source.inline,
                      resource_path=os.path.join(resource_path_prefix, resource_name),
@@ -261,17 +259,86 @@ def _inline_code_list_for(component: Component, resource_path_prefix: str, base_
       ]
     else:
       return []
+  # for custom blob write
+  elif isinstance(component, WriteConnector):
+    icl = []
+    if component.container:
+      details = component.container.record_connection.details
+      write_strategy_value = value_from_proto(details.get('write_strategy', UnionValue()))
+      custom_function = write_strategy_value.get('custom_function')
+      if custom_function:
+        custom_code = custom_function.get('code_interface', {}).get('write_all', {}).get("code_fingerprint_strategy",
+                                                                                         UnionValue()).get("content_based_fingerprint", {}).get("code")
+        if custom_code:
+          icl.append(
+              InlineCode(
+                  code=custom_code,
+                  resource_path=os.path.join(resource_path_prefix, f"{component.id}_custom_code.py"),  # noqa: E126
+                  base_path=base_path,
+                  attribute_path=('container', 'record_connection', 'details', 'write_strategy', 'custom_function', 'code_interface', 'write_all',
+                                  'code_fingerprint_strategy', 'content_based_fingerprint', 'code'),
+                  base64_encoded=False))
+    return icl
+  # TODO(xiaoxi): refactor to make this shorter by traverse down by following the attribute path after extracting the value_from_proto
   elif isinstance(component, ReadConnector):
     icl = []
-    if component.container and component.container.HasField("byte_function") and component.container.byte_function.container.executable.code.source.HasField(
-        "inline"):
-      icl.append(
-          InlineCode(
-              code=base64.b64decode(component.container.byte_function.container.executable.code.source.inline).decode(),
-              resource_path=os.path.join(resource_path_prefix, f"custom_read_connector_{component.id}.py"),  # noqa: E126
-              base_path=base_path,
-              attribute_path=("container", "byte_function", "container", "executable", "code", "source", "inline"),
-              base64_encoded=True))
+    if component.container:
+      if component.container.HasField("byte_function") and component.container.byte_function.container.executable.code.source.HasField("inline"):
+        icl.append(
+            InlineCode(
+                code=base64.b64decode(component.container.byte_function.container.executable.code.source.inline).decode(),
+                resource_path=os.path.join(resource_path_prefix, f"custom_read_connector_{component.id}.py"),  # noqa: E126
+                base_path=base_path,
+                attribute_path=("container", "byte_function", "container", "executable", "code", "source", "inline"),
+                base64_encoded=True))
+      elif component.container.HasField("record_connection"):
+        details = component.container.record_connection.details
+        code_interface = details.get("code_interface", {})
+        # new custom python RC
+        if code_interface:
+          code_interface_value = value_from_proto(code_interface)
+          code = code_interface_value.get("bytes", dict()).get("code_fingerprint_strategy", UnionValue()).get("content_based_fingerprint", {}).get("code")
+          if code:
+            icl.append(
+                InlineCode(
+                    code=code,
+                    resource_path=os.path.join(resource_path_prefix, f"{component.id}_custom_code.py"),  # noqa: E126
+                    base_path=base_path,
+                    attribute_path=('container', 'record_connection', 'details', 'code_interface', 'bytes', 'code_fingerprint_strategy',
+                                    'content_based_fingerprint', 'code'),
+                    base64_encoded=False))
+
+          parser_type = code_interface_value.get("bytes", dict()).get("parser_type", UnionValue())
+          if parser_type.get("user_parser", {}):
+            parser_code = parser_type.get("user_parser", {}).get("code_interface",
+                                                                 {}).get("byte_stream", {}).get("code_fingerprint_strategy",
+                                                                                                UnionValue()).get("content_based_fingerprint", {}).get("code")
+            if parser_code:
+              icl.append(
+                  InlineCode(
+                      code=parser_code,
+                      resource_path=os.path.join(resource_path_prefix, f"{component.id}_custom_parser.py"),  # noqa: E126
+                      base_path=base_path,
+                      attribute_path=("container", "record_connection", "details", "code_interface", "bytes", "parser_type", "user_parser", "code_interface",
+                                      "byte_stream", "code_fingerprint_strategy", "content_based_fingerprint", "code"),
+                      base64_encoded=False))
+        # for user parser in blob rc
+        parser = details.get('parser_type')
+        if parser:
+          parser_value = value_from_proto(parser)
+          parser_code = parser_value.get('user_parser', {}).get('code_interface', {}).get('byte_stream', {}).get('code_fingerprint_strategy',
+                                                                                                                 {}).get('content_based_fingerprint',
+                                                                                                                         {}).get('code')
+          if parser_code:
+            icl.append(
+                InlineCode(
+                    code=parser_code,
+                    resource_path=os.path.join(resource_path_prefix, f"{component.id}_custom_parser.py"),  # noqa: E126
+                    base_path=base_path,
+                    attribute_path=("container", "record_connection", "details", "parser_type", "user_parser", "code_interface", "byte_stream",
+                                    "code_fingerprint_strategy", "content_based_fingerprint", "code"),
+                    base64_encoded=False))
+
     if component.bytes and component.bytes.parser.HasField("lambda_parser") and component.bytes.parser.lambda_parser.HasField(
         "code") and component.bytes.parser.lambda_parser.code.HasField("inline"):
       icl.append(
@@ -372,13 +439,14 @@ def _render_map(mp, attribute_path: Tuple[str, ...], indent=0, classname_map: di
   tmpl = jinja2.Template('''{
 {% for k in sorted(mp.keys()) %}\
 {{_spaces_for(indent+2)}}{{_render_value(k, attribute_path, indent+2, classname_map, attribute_overrides)}}: \
-{{ _render_value(mp[k], attribute_path, indent+2, classname_map, attribute_overrides) }},\n\
+{{ _render_value(mp[k], _append(attribute_path, k), indent+2, classname_map, attribute_overrides) }},\n\
 {% endfor %}\
 {{_spaces_for(indent)}}}''')
   return tmpl.render(mp=mp,
                      sorted=sorted,
                      _render_value=_render_value,
                      _spaces_for=_spaces_for,
+                     _append=_append,
                      indent=indent,
                      classname_map=classname_map,
                      attribute_path=attribute_path,
@@ -390,6 +458,10 @@ def _render_proto_field(field: schema_pb2.Field, attribute_path: Tuple[str, ...]
   args = ''
   if kind == 'array':
     args += f', {_render_value(field.schema.array.element_schema, attribute_path, indent, classname_map, attribute_overrides)}'
+  elif kind == 'decimal':
+    precision = field.schema.decimal.precision or 10
+    scale = field.schema.decimal.scale or 0
+    args += f', {precision}, {scale}'
   elif kind == 'dictionary':
     # We wrap these in Field objects as adding in more wrappers (this time for Schema types)
     # seems like overkill.
@@ -408,11 +480,18 @@ def _render_proto_value(val: ascend_pb2.Value, attribute_path: Tuple[str, ...],
   if kind == 'struct_value':
     args = _render_value(val.struct_value.fields, attribute_path, indent, classname_map, attribute_overrides)
   elif kind == 'union_value':
-    args = f"'{val.union_value.tag}'"
+    fields = '{}'
     if val.union_value.value.struct_value.fields:
-      args += f', {_render_value(val.union_value.value, attribute_path, indent, classname_map, attribute_overrides)}'
+      fields = _render_value(val.union_value.value.struct_value.fields, _append(attribute_path, val.union_value.tag), indent, classname_map,
+                             attribute_overrides)
+    args = f"{val.union_value.tag}={fields}"
   else:
-    args = _render_value(getattr(val, kind), attribute_path, indent, classname_map, attribute_overrides)
+    if attribute_overrides.get(attribute_path) and attribute_path[-1] == 'code':
+      tmpl = jinja2.Template('''{{attribute_overrides.get(attribute_path)}}''')
+      t = ''.join(w.capitalize() for w in kind.split('_')[:-1])
+      return f"value.{t}({tmpl.render(attribute_overrides=attribute_overrides, attribute_path=attribute_path)})"
+    else:
+      args = _render_value(getattr(val, kind), attribute_path, indent, classname_map, attribute_overrides)
   t = ''.join(w.capitalize() for w in kind.split('_')[:-1])
   return f"value.{t}({args})"
 
